# filename: quads.py

'''
Purpose:
    Provides input to SST.
'''

# ===== define globals =====
DEBUG = True
#DEBUG = False

# ===== Global Imports =====
import sys
errflagMissingModule = False
stringMissingModule = ''

try:
    import sst
except:
    errflagMissingModule = True
    stringMissingModule += (
        'Cannot import sst module.\n'
        )

try:
    from optparse import OptionParser
except:
    errflagMissingModule = True
    stringMissingModule += (
        'Cannot import optparse module.\n'
        )

if errflagMissingModule:
    print("\n" + stringMissingModule)
    print("\nStopping in file quads.py\n")
    sys.exit()
else:
    if DEBUG:
        print("\nAll modules imported correctly.\n")


# = = = = = classes = = = = = = = = =
class RtrPorts:
    def __init__(self):
        self._next_addr = 0 
    def nextPort(self):
        res = self._next_addr
        self._next_addr = self._next_addr + 1
        return res
    def numPorts(self):
        if DEBUG: 
            print('RtrPorts: self._next_addr = %s' % self._next_addr)
        return self._next_addr


# = = = = = functions = = = = = = = = 

def doQuad(quad, cores):
    if DEBUG:
        print('\n\n*** In function doQuad ***\n')

    sst.pushNamePrefix("q%d"%quad)

    bus = sst.Component("membus", "memHierarchy.Bus")
    bus.addParams({
        "bus_frequency" : clock,
        "bus_latency_cycles" : 1,
        })

    if DEBUG:
        print
        print("Prefix: q%d" % quad)
        print("bus_frequency: %s" % clock)
        print("bus_latency_cycles: %s" % 1)
        print

    for x in range(cores):
        core = 4*quad + x
       
        if DEBUG:
            print
            print(30*'=')
            print('> FOR CORE %s:' % core)

        l1id = sst.Component("l1cache_%d"%core, "memHierarchy.Cache")
        dictParamsL1id = {
            "coherence_protocol": coherence_protocol,
            "cache_frequency": clock,
            "replacement_policy": "lru",
#            "cache_size": "16KB",
            "cache_size": "1KB",
            "associativity": 8,
            "cache_line_size": 64,
            "low_network_links": 1,
            "access_latency_cycles": 2,
            "L1": 1,
            "statistics": 1,
            "debug": memDebug,
            "debug_level" : 6,
            }
        if DEBUG:
            print
            print(' === l1id ===')
            for key in dictParamsL1id.keys():
                print(' %s: %s' % (key,dictParamsL1id[key]))
            print
        l1id.addParams(dictParamsL1id)

        if DEBUG:
            print
            print(' === l1PrefetchParams ===')
            print('%s' % l1PrefetchParams)
            print
        l1id.addParams(l1PrefetchParams)

        arielL1Link = sst.Link("cpu_cache_link_%d"%core)
        arielL1Link.connect((ariel, "cache_link_%d"%core, busLat), (l1id, "high_network_0", busLat))
        membusLink = sst.Link("cache_bus_link_%d"%core)
        membusLink.connect((l1id, "low_network_0", busLat), (bus, "high_network_%d"%x, busLat))

    l2 = sst.Component("l2cache", "memHierarchy.Cache")
    dictParamsL2 = {
        "coherence_protocol": coherence_protocol,
        "cache_frequency": clock,
        "replacement_policy": "lru",
        "cache_size": "512KB",
        "associativity": 16,
        "cache_line_size": 64,
        "access_latency_cycles": 23,
        "low_network_links": 1,
        "high_network_links": 1,
        "mshr_num_entries" : 128,
        "L1": 0,
        "directory_at_next_level": 1,
        "network_address": quad,
        "network_bw": coreNetBW,
        "statistics": 1,
        "debug_level" : 6,
        "debug": memDebug
        }
    if DEBUG:
        print
        print(30*'=')
        print(" === l2 ===")
        for key in dictParamsL2.keys():
            print('%s: %s' % (key,dictParamsL2[key]))
        print
    l2.addParams(dictParamsL2)

    if DEBUG:
        print
        print(30*'=')
        print(" === l2PrefetchParams === ")
        print(l2PrefetchParams)
        print

    l2.addParams(l2PrefetchParams)

    link = sst.Link("l2cache_%d_link"%quad)
    link.connect((l2, "high_network_0", busLat), (bus, "low_network_0", busLat))
    link = sst.Link("l2cache_%d_netlink"%quad)
    link.connect((l2, "directory", netLat), (rtr, "port%d"%(rtrInfo.nextPort()), netLat))

    sst.popNamePrefix()


def doMem(id, memParams, memclock, num_channels, interleave_size, rangeStart):
    if DEBUG:
        print('\n\n*** In function doMem ***\n')
    rank_size = 512 / num_channels;
    start_pos = rangeStart;
    interleave_step = num_channels*(interleave_size/1024) # in KB
    end_pos = rangeStart + ((512*1024*1024)-(interleave_size*(num_channels-1))) 

    for chan in range(num_channels) :

        if DEBUG:
            print
            print(30*"+")
            print('> FOR CHANNEL %s:' % chan)

        sst.pushNamePrefix("mem%d.%02d"%(id,chan))

        if DEBUG:
            print(' Prefix: mem%d.%02d' % (id,chan))

        memory = sst.Component("memory", "memHierarchy.MemController")

        dictMemory = {
            "coherence_protocol": coherence_protocol,
            "rangeStart": 0,
            "mem_size": rank_size,
            "clock": memclock,
            "statistics": 1,
            "debug": memDebug
            }
        if DEBUG:
            print
            print(' === memory ===')
            for key in dictMemory.keys():
                print(' %s: %s' % (key, dictMemory[key]))
            print
        memory.addParams(dictMemory)

        if DEBUG:
            print
            print(' === memParams ===')
            print('%s' % memParams)
        memory.addParams(memParams)

        netPort = rtrInfo.nextPort()
        dc = sst.Component("dc", "memHierarchy.DirectoryController")

        dictDC = {
            "coherence_protocol": coherence_protocol,
            "network_bw": memNetBW,
            "addr_range_start": start_pos,
            "addr_range_end":  end_pos,
            "interleave_size": interleave_size/1024,    # in KB
            "interleave_step": interleave_step,         # in KB
            "entry_cache_size": 128*1024, #Entry cache size of mem/blocksize
            "clock": memclock,
            "debug": memDebug,
            "statistics": 1,
            "network_address": netPort
            }
        if DEBUG:
            print
            print(" === dc ===")
            for key in dictDC.keys():
                print(' %s: %s' % (key,dictDC[key]))
            print
        dc.addParams(dictDC)

        memLink = sst.Link("mem%d_link"%id)
        memLink.connect((memory, "direct_link", busLat), (dc, "memory", busLat))
        netLink = sst.Link("dc%d_netlink"%id)
        netLink.connect((dc, "network", netLat), (rtr, "port%d"%(netPort), netLat))

        sst.popNamePrefix()
        start_pos = start_pos + interleave_size
        end_pos = end_pos + interleave_size
        if DEBUG:
            print
            print("start_pos: %s" % start_pos)
            print("end_pos: %s" % end_pos)
            print

# = = =  = end of functions = = = = = 

# = = = = = = main = = = = = = = = = 

if DEBUG:
    print("\n\n ====== quads.py ======\n")

rtrInfo = RtrPorts()

memDebug = 0
memDebugLevel = 7
baseclock = 1500  # in MHz
clock = "%g MHz"%(baseclock)
l2clock = "%g MHz"%(baseclock *0.25)
coherence_protocol = "MESI"
busLat = "50ps"
netLat = "3ns"
coreNetBW = "72GB/s"
memNetBW = "36GB/s"
xbarBW = coreNetBW
flit_size = "72B"
# @ 16-byte flit, 2GHz clock
# @ 32-byte flit, 1GHz clock

if DEBUG:
    print
    print("memDbug: %s" % memDebug)
    print("memDebugLevel: %s" % memDebugLevel)
    print("baseclock: %s MHz" % baseclock)
    print("clock: %s" % clock)
    print("l2clock: %s" % l2clock)
    print("coherence_protocol: %s" % coherence_protocol)
    print("busLat: %s" % busLat)
    print("netLat: %s" % netLat)
    print("coreNetBW: %s" % coreNetBW)
    print("memNetBW: %s" % memNetBW)
    print("xbarBW: %s" % xbarBW)
    print("flit_size: %s" % flit_size)
    print

l1PrefetchParams = { }

dictL2PrefetchParams = {
        "prefetcher": "cassini.StridePrefetcher",
        "reach": 8
        }
if DEBUG:
    print
    print(" === l2PrefetchParams ===")
    for key in dictL2PrefetchParams.keys():
      print(" %s: %s" % (key,dictL2PrefetchParams[key]))
    print
l2PrefetchParams = (dictL2PrefetchParams)


op = OptionParser()
# 1 = qsorting, 2 = std::sort, 3 = __gnu_parallel::parallel_sort_mwms, 4 = xgc_sort
op.add_option("-n", "--num_elems", action="store", type="int", dest="nelem", default=1000000)
op.add_option("-c", "--num_cores", action="store", type="int", dest="cores", default=8)
op.add_option("-d", "--memDebug", action="store", type="int", dest="memDebug", default=0)
op.add_option("-x", "--executable", action="store", type="string", dest="testExecutable", default="/home/dwbarne/local/sst-4.0/bin/smac2d")

# type(op.parse_args()) is tuple
(options, args) = op.parse_args()
if DEBUG:
    print("\n === options from op.parse_args() ===")
    print("  --num_elems: %s" % options.nelem)
    print("  --num_cores: %s" % options.cores)
    print("  --memDebug: %s" % options.memDebug)
    print("  --executable: %s" % options.testExecutable)
    print 

memDebug = options.memDebug
corecount = options.cores
quads = corecount / 4
if corecount % 4:
    quads = quads+1
    if (DEBUG):
        print(" quads: %s" % quads)
        print


ariel = sst.Component("a0", "ariel.ariel")
dictAriel = {
	"verbose" : 0,
        "clock" : clock,
	"maxcorequeue" : 256,
 	"maxissuepercycle" : 2,
 	"pipetimeout" : 0,
	"corecount" : corecount,
 	"arielmode" : 1,
        "arieltool": "/home/dwbarne/scratch/src/sst-4.0/tools/ariel/fesimple/fesimple.so",
        "appargcount" : 1,
        "apparg0" : corecount,
        "executable": options.testExecutable,
 	"memorylevels" : 1,
	"defaultlevel" : 0
    }
if (DEBUG):
    for key in dictAriel.keys():
        print("%s: %s" % (key,dictAriel[key]))
    print
ariel.addParams(dictAriel)


# Build network
rtr = sst.Component("chipRtr", "merlin.hr_router")
dictRtr = {
    "id": 0,
    "topology": "merlin.singlerouter",
    "link_bw": coreNetBW,
    "xbar_bw": xbarBW,
    "input_latency": "6ns",
    "output_latency": "6ns",
    "input_buf_size" : "4KiB",
    "output_buf_size" : "4KiB",
    "flit_size" : flit_size
    }
if DEBUG:
    for key in dictRtr.keys():
        print("%s: %s" % (key,dictRtr[key]))
    print
rtr.addParams(dictRtr)


for x in range(quads-1):
    doQuad(x, 4)
doQuad(quads-1, corecount-(4*(quads-1)))

dictMemParams = {
        "access_time" : "70ns"
        #"backend" : "memHierarchy.dramsim",
        #"device_ini" : "DDR3_micron_64M_8B_x16_sg1.ini"
        #"system_ini" : "system-ddr3.ini"
        }
if DEBUG:
    print
    print(" === dictMemParams === ")
    for key in dictMemParams.keys():
        print("%s: %s" % (key,dictMemParams[key]))
    print

doMem(0, dictMemParams, "250MHz", 4, 1024, 0) # 64GB/s
    # 4 channels -> Interleave  [ 4 * 250MHz ]

rtr.addParam("num_ports", rtrInfo.numPorts())

print "\n>> done configuring SST <<\n"


# = = = = = = end of main = = = = = = = = = 

